def XWIN_single_instance(model, model_type, mode, df, metadata_json,  predictors_list=None, use_imputer=False, DEBUG=False):

    '''
    XWIN_single_instance

    Parameters
    ----------
    - model, whose predictions are to be explained
    - model_type: 'sklearn' or 'model_ops'
    - mode ('classification' or 'regression', to select predict_proba or predict)
    - df, instance for which a prediction is made
    - predictors (not all columns in df need to be predictors, maybe default should be df.columns)
    - metadata_json
    - use_imputer = False [default] or True. If True, the trained imputer imputes the withheld information, rather than using the metadata.
      Note: in the True case metadata_json is still used to determine which columns are numeric
    - DEBUG = True for print statements

    output:
    - XWIN_impact_table, table whose columns are: Feature,
                                                  XWIN Impact Value,
                                                  Order,
                                                  Relative Feature Value,
                                                  Explanation In Words
    - XWIN_impact_table is sorted by key=Order

   '''


    # Imports
    # =======

    import cloudpickle
    import json
    import numpy as np
    import pandas as pd
    from spotfire_dsml.ml_metadata import metadata_data as mdd
    from spotfire_dsml.ml_metadata import metadata_model as mdm


    # Initializations
    # ===============


    #predictors_list = predictors.split(',')  # <<<<<<<<<<<<<<<< depends on format in which 'predictors' is passed in
    #dfX = df[predictors_list]
    if predictors_list is None:
        predictors_list = df.columns
    dfX = df[predictors_list]
    features = dfX.columns


    # Pick up mean and std and dtypes from the metadata
    # =================================================

    # assuming the metadata is passed as dict, first convert to string here
    # NOTE the new function uses a dict so I am commenting this
    # metadata_json = json.dumps(metadata_json) # dict to str

    received_desc_worker, df_meta = mdd.DescribeWorker.unpack_json(metadata_json, return_df=True)
    stats_df = received_desc_worker.get_metric(['mean', 'std', 'dtype']).set_index('variable')

    if DEBUG:
        print('Metadata:\n', stats_df, flush=True)

    all_numeric_columns = stats_df[stats_df['dtype'] != 'object'].index.values

    if DEBUG:
        print('All numeric columns:\n', all_numeric_columns, flush=True)

    numeric_columns = []
    for col in all_numeric_columns:
        if col in predictors_list:
            numeric_columns.append(col)

    if DEBUG:
        print('All numeric predictors:\n', numeric_columns, flush=True)


    # Create the auxiliary instances in one matrix with the actual instance
    # =====================================================================

    dfX = dfX.reset_index(drop=True)
    K = dfX.shape[1]

    # Initialize dfX_auxiliary in its correct shape, with all numeric columns in float64.
    # The reason for float64 is that by sending in NaNs the type would change to float64, and
    # potentially the memory for the matrix would have to be reallocated if the type is smaller than 64 bits.


    dfX_original_plus_auxiliary = dfX.loc[dfX.index.repeat(K+1)].reset_index(drop=True)

    for col in numeric_columns:
        dfX_original_plus_auxiliary[col] = dfX_original_plus_auxiliary[col].astype('float64')

    # Now we have the memory, we assign the rows in dfX_auxiliary one at a time.
    # First row is the original instances, the next K rows are instances in which
    # one of the K features is modified. We here modify by substitution
    # with the mean values, but substituting with NaN is an option for later.

    dfX_original_plus_auxiliary.iloc[0] = dfX.iloc[0].copy()

    # Loop over all features and do the substitution
    j = 0
    for feature in features:
        if DEBUG:
            print(feature, flush=True)
        if feature in numeric_columns:
            if not use_imputer:
                dfX_original_plus_auxiliary.loc[j + 1,feature] = stats_df.loc[feature,'mean']
            else:
                dfX_original_plus_auxiliary.loc[j + 1,feature] = np.nan
        else:
            dfX_original_plus_auxiliary.loc[j + 1,feature] = '__UNKNOWN_CATEGORY__'
        j += 1

    # Compute the predictions for the entire group of auxiliary instances and the original instance
    # =============================================================================================

    scorer = mdm.init_scorer(model_type, model, mode)

    if mode =='classification':
        predictions = scorer.predict_probability(dfX_original_plus_auxiliary)[:,1]
    elif mode == 'regression':
        predictions = scorer.predict(dfX_original_plus_auxiliary)

    # Compute the impact values
    # =========================
    values = (predictions[0] - predictions)[1:]
    shifts = dict(zip(features, values))
    sorted_shifts_dict = dict(sorted(shifts.items(), key=lambda x: np.abs(x[1]), reverse=True))
    keys = list(sorted_shifts_dict.keys())
    values = list(sorted_shifts_dict.values())


    # Compute the relative feature values
    # ===================================

    relative_feature_values = []
    for key,val in zip(keys,values):
        if key not in numeric_columns:
            relative_feature_values.append(0.0)
            continue
        mean_key = stats_df.loc[key,'mean']
        std_key = stats_df.loc[key,'std']
        if np.isnan(dfX.loc[0,key]):
            relative_feature_value = 0.0
        else:
            relative_feature_value = (dfX.loc[0,key] - mean_key)/std_key
        if relative_feature_value > 2.0:
            relative_feature_value = 2.0
        elif relative_feature_value < -2.0:
            relative_feature_value = -2.0
        relative_feature_value /= 2.0
        relative_feature_values.append(relative_feature_value)


    # Create the XWIN impact table
    # ============================

    XWIN_impact_table = pd.DataFrame.from_dict({'Feature':keys, 'XWIN Impact Value':values})
    XWIN_impact_table['Relative Feature Value'] = relative_feature_values
    XWIN_impact_table['Importance Order'] = np.arange(1, XWIN_impact_table.shape[0] + 1)


    # Add explanations_in_words
    # =========================

    explanations_in_words = []
    for i, feature in enumerate(XWIN_impact_table['Feature'].values):
        relative_value_in_words = 'near-average'
        if XWIN_impact_table.loc[len(explanations_in_words), 'Relative Feature Value'] > 0.99:
            relative_value_in_words = 'very high'
        elif XWIN_impact_table.loc[len(explanations_in_words), 'Relative Feature Value'] > 0.5:
            relative_value_in_words = 'high'
        elif XWIN_impact_table.loc[len(explanations_in_words), 'Relative Feature Value'] > 0.25:
            relative_value_in_words = 'moderately high'
        elif XWIN_impact_table.loc[len(explanations_in_words), 'Relative Feature Value'] > 0.05:
            relative_value_in_words = 'somewhat elevated'
        elif XWIN_impact_table.loc[len(explanations_in_words), 'Relative Feature Value'] < -0.99:
            relative_value_in_words = 'very low'
        elif XWIN_impact_table.loc[len(explanations_in_words), 'Relative Feature Value'] < -0.5:
            relative_value_in_words = 'low'
        elif XWIN_impact_table.loc[len(explanations_in_words), 'Relative Feature Value'] < -0.25:
            relative_value_in_words = 'moderately low'
        elif XWIN_impact_table.loc[len(explanations_in_words), 'Relative Feature Value'] < -0.05:
            relative_value_in_words = 'somewhat depressed'

        amount = XWIN_impact_table.loc[len(explanations_in_words), 'XWIN Impact Value']
        if amount >= 0:
            direction = 'increases'
        elif amount < 0:
            direction = 'decreases'
        amount = str(round(amount,2))
        explanations_in_words.append('The ' + relative_value_in_words + ' value of ' + feature + ' ' + direction + ' the model-prediction by ' + amount)

    XWIN_impact_table['Explanation In Words'] = explanations_in_words

    if DEBUG:
        print('XWIN_impact_table:\n', XWIN_impact_table, flush=True)

    return XWIN_impact_table


def XWIN_batch(model, model_type, mode, df, metadata_json, predictors_list=None, use_imputer=False, DEBUG=False):

    '''
    XWIN_batch

    Parameters
    ----------
    - model, whose predictions are to be explained
    - model_type: 'sklearn' or 'model_ops'
    - mode (classification or regression, to select predict_proba or predict)
    - df, batch of instances for which a predictions are made
    - predictors (not all columns in df need to be predictors, maybe default should be df.columns)
    - metadata_json
    # - use_imputer = False [default] or True. If True, the trained imputer imputes the withheld information, rather than using the metadata.
    #   Note: in the True case metadata_json is still used to determine which columns are numeric, and for computing Relative Feature Value
    # - DEBUG = True for print statements

    output:
    - XWIN_batch_impact_table, dataframe whose columns are: Feature,
                                                      XWIN Impact Value,
                                                      Order,
                                                      Relative Feature Value,
                                                      Explanation In Words
    - XWIN_batch_impact_table is sorted by key=Order

    '''


    # Imports
    # =======

    import cloudpickle
    import numpy as np
    import pandas as pd
    import json
    from spotfire_dsml.ml_metadata import metadata_data as mdd
    from spotfire_dsml.ml_metadata import metadata_model as mdm


    # Initializations
    # ===============

    #GAIA17May: This should go too as the scorer does the job now
    #model = cloudpickle.loads(model) # <<<<<<<<<<<<<<<< depends on how the model is passed in


    # Randomly select a batch of 'batch_size' instances
    # =================================================

    batch_size = 100
    #predictors_list = predictors.split(',')   # <<<<<<<<<<<<<<<< depends on format in which 'predictors' is passed in
    #dfX = df[predictors_list]
    if predictors_list is None:
        predictors_list = df.columns
    dfX = df[predictors_list]

    if dfX.shape[0] > batch_size:
        n_explanation_instances = batch_size
        scorer = mdm.init_scorer('sklearn', model, mode)
        half_batch_size = int(batch_size/2)
        if mode == 'regression':
            preds = scorer.predict(dfX)
            # take equal sample of small and large predictions, say two outer quartiles
            Q1 = np.quantile(preds, 0.25)
            Q3 = np.quantile(preds, 0.75)
            large_preds_idx = np.where(preds > Q3)
            small_preds_idx = np.where(preds < Q1)
            np.random.seed(59)        
            if len(large_preds_idx) > half_batch_size:
                large_preds_idx_half_batch_size = np.random.choice(large_preds_idx, half_batch_size, replace=False)
            else:
                large_preds_idx_half_batch_size = large_preds_idx
            if len(small_preds_idx) > half_batch_size:
                small_preds_idx_half_batch_size = np.random.choice(small_preds_idx, half_batch_size, replace=False)
            else:
                small_preds_idx_half_batch_size = small_preds_idx
            idx = np.concatenate((large_preds_idx_half_batch_size, small_preds_idx_half_batch_size), axis=1)[0]
            dfX_batch = dfX.iloc[idx,:]      
        elif mode == 'classification':
            probs = scorer.predict_probability(dfX)[:,1]
            # take equal sample of small and large probabilities:
            large_probs_idx = np.where(probs > 0.8)[0]
            small_probs_idx = np.where(probs < 0.2)[0]
            np.random.seed(59)
            if len(large_probs_idx) > half_batch_size:
                large_probs_idx_half_batch_size = np.random.choice(large_probs_idx, half_batch_size, replace=False)
            else:
                large_probs_idx_half_batch_size = large_probs_idx
            if len(small_probs_idx) > half_batch_size:
                small_probs_idx_half_batch_size = np.random.choice(small_probs_idx, half_batch_size, replace=False)
            else:
                small_probs_idx_half_batch_size = small_probs_idx    
            idx = np.concatenate((large_probs_idx_half_batch_size, small_probs_idx_half_batch_size))
            dfX_batch = dfX.iloc[idx,:] 
    else:
        n_explanation_instances = dfX.shape[0]
        dfX_batch = dfX
        
    batch_size = dfX_batch.shape[0]

    if DEBUG:
        print('Batch info:\n', dfX_batch.info(), flush=True)


    # Pick up mean and std and dtypes from the metadata
    # =================================================

    # assuming the metadata is passed as dict, first convert to string here
    # NOTE the new function uses a dict so I am commenting this
    # metadata_json = json.dumps(metadata_json) # dict to str

    received_desc_worker, df_meta = mdd.DescribeWorker.unpack_json(metadata_json, return_df=True)
    stats_df = received_desc_worker.get_metric(['mean', 'std', 'dtype']).set_index('variable')

    if DEBUG:
        print('Metadata:\n', stats_df, flush=True)

    all_numeric_columns = stats_df[stats_df['dtype'] != 'object'].index.values

    if DEBUG:
        print('All numeric columns:\n', all_numeric_columns, flush=True)

    numeric_columns = []
    for col in all_numeric_columns:
        if col in predictors_list:
            numeric_columns.append(col)

    if DEBUG:
        print('All numeric predictors:\n', numeric_columns, flush=True)


    # Compute data for additional information: relative feature values
    # (the colors of the datapoints in the scatter plot)
    # ================================================================

    color_matrix = np.zeros(dfX_batch.shape)
    for i in range(color_matrix.shape[0]):
        for j in range(color_matrix.shape[1]):
            if dfX.columns[j] in numeric_columns:
                #if DEBUG:
                #    print('Numeric feature = ', dfX.columns[j], ' with value ', dfX_batch.iloc[i,j], flush=True)
                if np.isnan(dfX_batch.iloc[i,j]):
                    color_matrix[i,j] = 0
                else:
                    color_matrix[i,j] = (dfX_batch.iloc[i,j] - stats_df.loc[dfX.columns[j],'mean'])/stats_df.loc[dfX.columns[j],'std']
            else:
                #if DEBUG:
                #    print('Categoric feature = ', dfX.columns[j], ' with value ', dfX_batch.iloc[i,j], flush=True)
                color_matrix[i,j] = 0
            # Do truncation for meaningful colors that are not determined by the outliers
            if color_matrix[i,j] > 2.0:
                color_matrix[i,j] = 2.0
            if color_matrix[i,j] < -2.0:
                color_matrix[i,j] = -2.0
            color_matrix[i,j] /= 2.0


    # Compute the impact values for all features of all instances in the batch
    # Implementation: use only one call to the predict service, anticipating
    # a call to the ModelOps API for scoring.
    # ========================================================================

    features = dfX.columns
    impact_matrix = np.zeros(dfX_batch.shape)
    numeric_columns = dfX_batch.select_dtypes(exclude='object').columns
    categoric_columns = dfX_batch.select_dtypes(include='object').columns
    if DEBUG:
        print('impact matrix shape:', impact_matrix.shape, flush=True)

    # Create dfX_batch_auxiliary that will be sent to the predict service to get all auxiliary predictions
    # ----------------------------------------------------------------------------------------------------

    K = dfX_batch.shape[1]

    # Initialize dfX_batch_auxiliary in its correct shape, with all numeric columns in float64.
    # The reason for float64 is that by sending in NaNs the type would change to float64, and
    # potentially the memory for the matrix would have to be reallocated if the type is smaller than 64 bits.

    dfX_batch_auxiliary = dfX_batch.loc[dfX_batch.index.repeat(K+1)].reset_index(drop=True)
    for col in numeric_columns:
        dfX_batch_auxiliary[col] = dfX_batch_auxiliary[col].astype('float64')

    # Now we have the memory, we assign the rows in dfX_batch_auxiliary one at a time.
    # Layout: 'batch_size' instances give 'batch_size' groups, each group consisting of K + 1
    # rows where K is the number of features. Each group first has the original row, followed
    # by the K rows that are the modified versions of that row. We here modify by substitution
    # with the mean values, but substituting with NaN is an option for later.

    # Loop over the original rows in the batch
    for i in range(dfX_batch.shape[0]):

        # The first row in each group in dfX_batch_auxiliary is the original row
        dfX_batch_auxiliary.iloc[i * (K + 1)] = dfX_batch.iloc[i].copy()

        # For each original row loop over all features and do the sunstitution
        j = 0
        for feature in features:
            if feature in numeric_columns:
                if not use_imputer:
                    dfX_batch_auxiliary.loc[i * (K + 1) + (j + 1),feature] = stats_df.loc[feature,'mean']
                else:
                    dfX_batch_auxiliary.loc[i * (K + 1) + (j + 1),feature] = np.nan
            else:
                dfX_batch_auxiliary.loc[i * (K + 1) + (j + 1),feature] = '__UNKNOWN_CATEGORY__'

            j += 1

    # Send dfX_batch_auxiliary for scoring
    # ------------------------------------

    scorer = mdm.init_scorer(model_type, model, mode)
    if mode =='classification':
        auxiliary_predictions = scorer.predict_probability(dfX_batch_auxiliary)[:,1]
    elif mode == 'regression':
        auxiliary_predictions = scorer.predict(dfX_batch_auxiliary)

    # Construct the impact matrix itself
    # ----------------------------------

    impact_matrix = np.zeros((batch_size, K))

    # For every row in the original batch ...
    for i in range(batch_size):
        # ... compute the impact values from the relevant auxiliary_predictions components
        impact_values = auxiliary_predictions[i * (K + 1)] - auxiliary_predictions[i * (K + 1) + 1: (i + 1) * (K + 1)]
        impact_matrix[i,:] = impact_values

    # Combine the impact values and the relative feature values in one dataframe
    # ==========================================================================

    df_out_xwin = pd.DataFrame(data=impact_matrix.transpose(), index=dfX_batch.columns)
    df_out_xwin['sortkey'] = np.sum(np.abs(df_out_xwin), axis=1)
    df_colors = pd.DataFrame(color_matrix.transpose(), index=df_out_xwin.index)
    df_out_xwin = pd.concat([df_out_xwin, df_colors], axis=1)


    # Create output in a format that is digestible for Spotfire (or other visualization tools)
    # ========================================================================================

    # sort
    df_out_xwin = df_out_xwin.sort_values(by='sortkey', ascending=False).drop('sortkey', axis=1)
    df_out_xwin.insert(0, 'Feature order', np.arange(1, df_out_xwin.shape[0] + 1))
    df_out_xwin.reset_index(inplace=True)
    df_out_xwin.rename(columns={'index':'Feature'}, inplace=True)

    # output digestible by Spotfire
    df_1 = df_out_xwin.iloc[:,:2+n_explanation_instances].copy()
    df_1_unpivot = df_1.melt(id_vars=['Feature','Feature order'], value_vars=df_1.columns[2:], var_name='Instance', value_name='Impact')
    df_2 = df_out_xwin.iloc[:,2+n_explanation_instances:].copy()
    df_2.insert(0, 'Feature', df_out_xwin['Feature'].copy())
    df_2_unpivot = df_2.melt(id_vars=['Feature'], value_vars=df_2.columns[1:], var_name='Instance', value_name='Relative FEATURE-value')
    df_xwin_batch_impact_table = df_1_unpivot
    df_xwin_batch_impact_table['Relative FEATURE-value'] = df_2_unpivot['Relative FEATURE-value']

    if DEBUG:
        print('df_xwin_batch_impact_table:\n', df_xwin_batch_impact_table, flush=True)

    return df_xwin_batch_impact_table




import pandas as _pd
import numpy as _np
import warnings as _warnings
from datetime import datetime as _datetime
from pandas.api.types import is_numeric_dtype as _is_numeric_dtype

_formatwarning_orig = _warnings.formatwarning
_warnings.formatwarning = lambda message, category, filename, lineno, line=None: _formatwarning_orig(message, category, filename, lineno, line='')

def _numeric_imputation(col, method, optional_args):
    warn = optional_args.get("warning_threshold")
    if col.isna().sum() >= (len(col) * warn):
        _warnings.warn('Warning: column "{}" contains more than {}% NaNs. Please consider further evaluation prior to imputation.'.format(col.name, int(warn * 100)))
    if method in ("nearest", 0):
        return col.interpolate("nearest").ffill().bfill()
    elif method in ("constant", 1):
        return col.fillna(optional_args.get("numeric_constant"))
    elif method in ("previous", 2):
        return col.fillna(method="ffill")
    elif method in ("next", 3):
        return col.fillna(method="bfill")
    elif method in ("linear", 4):
        return col.interpolate(method="linear")
    elif method in ("mean", 5):
        return col.fillna(col.mean())
    elif method in ("mean_over_n_previous", 6):
        return col.fillna(col.rolling(optional_args.get("n_previous"), min_periods=1).mean().shift(1))
    elif method in ("median_over_n_previous", 7):
        return col.fillna(col.rolling(optional_args.get("n_previous"), min_periods=1).median().shift(1))
    else:
        raise Exception("numeric_method not recognized, please try again.")

def _non_numeric_nearest(col):
    factors = col.factorize()
    factors_with_nan = _pd.Series(factors[0], col.index).replace(-1, _np.nan)
    factors_interpolated = factors_with_nan.interpolate("nearest").ffill().bfill()
    return factors_interpolated.map(dict(enumerate(factors[1])))
    
def _non_numeric_imputation(col, method, optional_args):
    warn = optional_args.get("warning_threshold")
    if col.isna().sum() >= (len(col) * warn):
        _warnings.warn('Warning: column "{}" contains more than {}% NaNs. Please consider further evaluation prior to imputation.'.format(col.name, int(warn * 100)))
    if method in ("nearest", 0):
        return _non_numeric_nearest(col)
    elif method in ("constant", 1):
        return col.fillna(optional_args.get("non_numeric_constant"))
    elif method in ("previous", 2):
        return col.fillna(method="ffill")
    elif method in ("next", 3):
        return col.fillna(method="bfill")
    else:
        raise Exception("non_numeric_method not recognized, please try again.")
    
def missing_value_imputation(ts, index, numeric_method="previous", non_numeric_method="previous", 
                             numeric_constant=0, non_numeric_constant="N/A", warning_threshold=0.5,
                             n_previous=3):
    """
    Imputing time series for missing values.

    Parameters
    ----------
    ts : data frame
        A data frame where a column is the timestamp or index and all of the
        other columns are readings of variables at corresponding timestamps.
        The number of columns can be an arbitrary number. The values can be
        numeric or non-numeric such as categorical (dictionary) or string. Required.
    index : str, int
        The index/timestamp column: if string it is interpreted as the name
        of the column. Value of -1 indicates that there is no index column and
        the function will create the index column. Required.
    numeric_method : str, int
        Type of imputation for numeric variables. Default = "previous".
        Options:
        0 or “nearest”
        1 or “constant” (constant specified with “numeric_constant”)
        2 or “previous”
        3 or “next”
        4 or “linear”
        5 or “mean”
        6 or “mean_over_n_previous”
        7 or “median_over_n_previous”
    non_numeric_method : str, int
        Type of imputation for non-numeric variables. Default = "previous".
        Options:
        0 or “nearest”
        1 or “constant”
        2 or “previous”
        3 or “next”
    numeric_constant : int, float, numeric
        The numeric constant to be used when the type of numeric imputation 
        is set to “constant”. Ignored when “numeric_method” is not
        constant. Default = 0.
    non_numeric_constant : str
        The non-numeric constant to be used when the type of non-numeric
        imputation is set to “constant”. Ignored when “non_numeric_method”
        is not constant. Default = N/A.
    n_previous : int
        The number of past values to include in calculation when
        “mean_over_n_previous” or “median_over_n_previous” is set for the numeric
        imputation, ignored otherwise. Default = 3.
    warning_threshold : float
        A factor that describes the minimum fraction of missing values per column
        needed to invoke a warning to the user. Default = 0.5.

    Returns
    -------
    ts
        ts with missing values imputed, same dimensions.

    Examples
    --------
    >>> ex = pd.DataFrame({"Date": pd.date_range("2015-01-01", "2015-01-05"), 
    >>>                        "Temp": [80, np.nan, 95, 76, np.nan],
    >>>                        "Events": ["Fog", np.nan, np.nan, "Fog", np.nan]})
    >>> missing_value_imputation(ex, "Date")
    >>>     Date		Temp	Events
    >>> 0	2015-01-01	80.00	Fog
    >>> 1	2015-01-02	83.67	Fog
    >>> 2	2015-01-03	95.00	Fog
    >>> 3	2015-01-04	76.00	Fog
    >>> 4	2015-01-05	83.67	Fog
    >>> missing_value_imputation(ex, "Date", numeric_method="mean_over_n_previous",
    >>>                              non_numeric_method="constant", n_previous=2)
    >>>     Date		Temp	Events
    >>> 0	2015-01-01	80.0	Fog
    >>> 1	2015-01-02	80.0	N/A
    >>> 2	2015-01-03	95.0	N/A
    >>> 3	2015-01-04	76.0	Fog
    >>> 4	2015-01-05	85.5	N/A
    >>> missing_value_imputation(ex, "Date", numeric_method="linear",
    >>>                              non_numeric_method="nearest")
    >>>     Date		Temp	Events
    >>> 0	2015-01-01	80.0	Fog
    >>> 1	2015-01-02	87.5	Fog
    >>> 2	2015-01-03	95.0	Fog
    >>> 3	2015-01-04	76.0	Fog
    >>> 4	2015-01-05	76.0	Fog

    """
    ts = ts.copy()
    args = locals()
    if index is None or index == -1: #add index if none exists
        index = "index"
        ts[index] = _np.arange(len(ts))
    numeric_cols = ts.select_dtypes("number").columns #separate numeric
    non_numeric_cols = ts.select_dtypes(exclude="number").columns #separate non-numeric
    # apply imputation to numeric and non-numeric columns
    ts[numeric_cols] = ts[numeric_cols].apply(lambda x: _numeric_imputation(x, method=numeric_method, optional_args=args))
    ts[non_numeric_cols] = ts[non_numeric_cols].apply(lambda x: _non_numeric_imputation(x, method=non_numeric_method, optional_args=args))
    return ts

def resampling(ts, index, rule=None, n_samples=None, fill_method="previous", sort=False):
    """
    Resampling time series data.
    
    This function returns a copy of the input time series resampled with
    a new desired frequency, specified by either a datetime rule or number
    of samples. Indices/timestamps will regular and equally distanced from 
    each other.

    Parameters
    ----------
    ts : data frame
        A data frame where a column is the timestamp or index and all of the
        other columns are readings of variables at corresponding timestamps.
        The number of columns can be an arbitrary number. The values can be 
        numeric or non-numeric such as categorical (dictionary) or string. Required.
    index : str
        The index/timestamp column. Required.
    rule : str
        Rule that represents how and at what frequency the timestamps will be 
        resampled. Examples: ("12H": 12 hours), ("3D": 3 days), ("Y", year)
        Full list of frequency strings can be found at:
        https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#dateoffset-objects
    n_samples : int
        Number of samples desired at completion of resampling process. If this
        input is present, “rule” will be ignored.
    fill_method : str, int
        Determines what method to either fill missing values introduced by 
        upsampling or aggregate values from downsampling. Default = "previous".
        Options:
        0 or “nearest”
        1 or “linear”
        2 or “previous”
        3 or “next”
        “mean” (only available when n_samples < len(ts))
        “median” (only available when n_samples < len(ts))
        “sum” (only available when n_samples < len(ts))
    sort : boolean
        If True, the function will sort the rows by the index. Default = False.

    Returns
    -------
    ts
        resampled copy of ts with desired frequency.

    Examples
    --------
    >>> ex = pd.DataFrame({"Date": pd.date_range("2015-01-01", "2015-01-05"), 
    >>>                        "Temp": [80, np.nan, 95, 76, np.nan],
    >>>                        "Events": ["Fog", np.nan, np.nan, "Fog", np.nan]})
    >>> resampling(ex, "Date", rule="12H")
    >>>     Date			Temp	Events
    >>> 0	2015-01-01 00:00:00	80.0	Fog
    >>> 1	2015-01-01 12:00:00	80.0	Fog
    >>> 2	2015-01-02 00:00:00	NaN	NaN
    >>> 3	2015-01-02 12:00:00	NaN	NaN
    >>> 4	2015-01-03 00:00:00	95.0	NaN
    >>> 5	2015-01-03 12:00:00	95.0	NaN
    >>> 6	2015-01-04 00:00:00	76.0	Fog
    >>> 7	2015-01-04 12:00:00	76.0	Fog
    >>> 8	2015-01-05 00:00:00	NaN	NaN
    >>> resampling(ex, "Date", n_samples=4, fill_method="nearest")
    >>>     Date			Temp	Events
    >>> 0	2015-01-01 00:00:00	80.0	Fog
    >>> 1	2015-01-02 08:00:00	NaN	NaN
    >>> 2	2015-01-03 16:00:00	76.0	Fog
    >>> 3	2015-01-05 00:00:00	NaN	NaN
    >>> resampling(ex, "Date", n_samples=3, fill_method="mean")
    >>>     Date		Temp	Events
    >>> 0	2015-01-01	80.0	Fog
    >>> 1	2015-01-03	85.5	NaN
    >>> 2	2015-01-05	NaN	NaN

    """
    ts = ts.copy()
    agg_methods = ["sum", "mean", "median"]
    numeric_cols = ts.select_dtypes("number").columns
    non_numeric_cols = ts.select_dtypes(exclude="number").columns
    if sort:
        ts = ts.sort_values(index)
    if not _is_numeric_dtype(ts[index]):
        ts[index] = _pd.to_datetime(ts[index])
        ts.index = ts[index]
    if n_samples:
        if (n_samples > len(ts)) and (fill_method in agg_methods):
            raise Exception("Aggregation methods are not allowed for upsampling. Please use a different fill_method.")
        idx_min = min(ts.index)
        idx_max = max(ts.index)
        try:
            resampled_index = _pd.Series(_np.linspace(idx_min, idx_max, num=n_samples))
        except:
            resampled_index = _pd.to_datetime(_np.linspace(idx_min.value, idx_max.value, num=n_samples))
        resampled_index.name = index
        ts = ts.fillna(_np.inf)
        ts = ts.reindex(ts.index.union(resampled_index))
        if fill_method in ("previous", 2):
            ts = ts.fillna(method="pad")
        elif fill_method in ("next", 3):
            ts = ts.fillna(method="backfill")
        elif fill_method in agg_methods:
            ts = _resample_agg(ts, index, fill_method, resampled_index, numeric_cols, non_numeric_cols)
        else:
            if fill_method in ("nearest", 0):
                ts[numeric_cols] = ts[numeric_cols].infer_objects().apply(lambda x: x.interpolate("nearest"))
                ts[non_numeric_cols] = ts[non_numeric_cols].apply(lambda x: _non_numeric_nearest(x))
            elif fill_method in ("linear", 1):
                ts[numeric_cols] = ts[numeric_cols].infer_objects().apply(lambda x: x.interpolate("linear"))
                ts[non_numeric_cols] = ts[non_numeric_cols].apply(lambda x: _non_numeric_nearest(x))
            else:
                raise Exception("fill_method not recognized, please try again.")
    elif rule:
        if fill_method in agg_methods:
            ts_non_num = ts[non_numeric_cols].resample(rule).fillna("nearest")
            if fill_method == "sum":
                ts_num = ts[numeric_cols].resample(rule).sum()
            elif fill_method == "mean":
                ts_num = ts[numeric_cols].resample(rule).mean()
            elif fill_method == "median": 
                ts_num = ts[numeric_cols].resample(rule).median()
            else:
                raise Exception("fill_method not recognized, please try again.")
            ts = _pd.concat([ts_num, ts_non_num], axis=1)
        else:
            ts = ts.resample(rule)
            if fill_method == "previous":
                ts = ts.fillna(method="pad")
            elif fill_method == "next":
                ts = ts.fillna(method="backfill")
            elif fill_method == "nearest":
                ts = ts.fillna(method="nearest")
            elif fill_method == "linear":
                ts = ts.mean().interpolate("linear")
            else:
                raise Exception("fill_method not recognized, please try again.")
    else:
        raise Exception("No resampling method recognized, please try again.")
    ts = ts.replace(_np.inf, _np.nan)
    try:
        ts = ts.reindex(resampled_index)
    except:
        pass
    return ts.drop(index, axis=1, errors="ignore").reset_index()

def _resample_agg(ts, index, fill_method, resampled_index, numeric_cols, non_numeric_cols):
    ts.index.name = "resampled_index"
    ts = ts.reset_index()
    ts["group"] = ts["resampled_index"].apply(lambda x: _np.searchsorted(resampled_index, x, side="right"))
    group_summary_num = ts.replace(_np.inf, _np.nan).groupby("group")[numeric_cols].agg(fill_method).reset_index()
    if len(non_numeric_cols) > 0:
        group_summary_non_num = ts.groupby("group")[non_numeric_cols].agg(lambda x: _pd.Series.mode(x)[0])
    ts = ts[["resampled_index", "group"]].merge(group_summary_num, on="group")
    if len(non_numeric_cols) > 0:
        ts = ts.merge(group_summary_non_num, on="group")
    return ts.set_index("resampled_index").drop("group", axis=1)

def min_max_normalization(ts, new_min=0, new_max=1):
    """
    Min/Max Normalization of time series data.
    
    Parameters
    ----------
    ts : data frame
        A data frame where a column is the timestamp or index and all of the
        other columns are readings of variables at corresponding timestamps.
        The number of columns can be an arbitrary number. The values can be 
        numeric or non-numeric such as categorical (dictionary) or string. Required.
    new_min : int, float, numeric
        The new desired minimum. Default = 0.
    new_max : int, float, numeric
        The new desired maximum. Default = 1.
    
    Returns
    -------
    ts
        Copy of ts with the numeric variables mapped to the new range (Non-numeric
        columns will be remain unchanged).
    
    Examples
    --------
    >>> ex = pd.DataFrame({"Date": pd.date_range("2015-01-01", "2015-01-05"), 
    >>>                        "Temp": [80, np.nan, 95, 76, np.nan],
    >>>                        "Events": ["Fog", np.nan, np.nan, "Fog", np.nan]})
    >>> min_max_normalization(ex)
    >>>     Date		Temp		Events
    >>> 0	2015-01-01	0.210526	Fog
    >>> 1	2015-01-02	NaN		NaN
    >>> 2	2015-01-03	1.000000	NaN
    >>> 3	2015-01-04	0.000000	Fog
    >>> 4	2015-01-05	NaN		NaN
    >>> min_max_normalization(ex, -5, 5)
    >>>     Date		Temp		Events
    >>> 0	2015-01-01	-2.894737	Fog
    >>> 1	2015-01-02	NaN		NaN
    >>> 2	2015-01-03	5.000000	NaN
    >>> 3	2015-01-04	-5.000000	Fog
    >>> 4	2015-01-05	NaN		NaN
    """
    if new_min >= new_max:
        raise Exception("new_max must be greater than new_min. Please adjust and try again.")
    ts = ts.copy()
    non_constant_cols = ts.loc[:, (ts != ts.iloc[0]).any()].columns
    numeric_cols = ts.select_dtypes("number").columns
    include_cols = numeric_cols.intersection(non_constant_cols)
    ts[include_cols] = ts[include_cols].apply(lambda x: (((x - _np.nanmin(x)) / (_np.nanmax(x) - _np.nanmin(x))) * \
                                                          (new_max - new_min)) + new_min)
    return ts

def index_normalization(ts, index, new_min, new_max, dt_format):
    """
    Index Normalization of time series data.

    This function stretches, squeezes, or shifts the time index
    of a time series to fit into a desired range.
    
    Parameters
    ----------
    ts : data frame
        A data frame where a column is the timestamp and all of the
        other columns are readings of variables at corresponding timestamps.
        The number of columns can be an arbitrary number. The values can be 
        numeric or non-numeric such as categorical (dictionary) or string. Required.
    index : str
        The timestamp column. Required.
    new_min : str
        The new desired minimum timestamp.
    new_max : str
        The new desired maximum timestamp.
    dt_format : str
        The datetime format of new_min/new_max.
        Examples: ("%m/%d/%y": "3/1/98"), ("%Y-%m-%d": "1998-03-01"), ("%m/%d/%Y %H:%M:%S": "3/1/1998 4:00:00")
        Reference for format codes:
        https://docs.python.org/3/library/datetime.html#strftime-and-strptime-format-codes

    Returns
    -------
    ts
        Copy of ts with the index mapped to its new range.

    Examples
    --------
    >>> ex = pd.DataFrame({"Date": pd.date_range("2015-01-01", "2015-01-05"), 
    >>>                        "Temp": [80, np.nan, 95, 76, np.nan],
    >>>                        "Events": ["Fog", np.nan, np.nan, "Fog", np.nan]})
    >>> index_normalization(ex, "Date", "2015-01-01", "2016-01-01", dt_format="%Y-%m-%d")
    >>>     Date			Temp	Events
    >>> 0	2015-01-01 00:00:00	80.0	Fog
    >>> 1	2015-04-02 06:00:00	NaN	NaN
    >>> 2	2015-07-02 12:00:00	95.0	NaN
    >>> 3	2015-10-01 18:00:00	76.0	Fog
    >>> 4	2016-01-01 00:00:00	NaN	NaN
    >>> index_normalization(ex, "Date", "12-31-14", "1-2-15", dt_format="%m-%d-%y")
    >>>     Date			Temp	Events
    >>> 0	2014-12-31 00:00:00	80.0	Fog
    >>> 1	2014-12-31 12:00:00	NaN	NaN
    >>> 2	2015-01-01 00:00:00	95.0	NaN
    >>> 3	2015-01-01 12:00:00	76.0	Fog
    >>> 4	2015-01-02 00:00:00	NaN	NaN

    """
    ts = ts.copy()
    new_min = _datetime.strptime(new_min, dt_format)
    new_max = _datetime.strptime(new_max, dt_format)
    x = _pd.to_datetime(ts[index])
    ts[index] = (((x - _np.min(x)) / (_np.max(x) - _np.min(x))) * (new_max - new_min)) + new_min
    return ts


# Copyright © 2023. Cloud Software Group, Inc.
# This file is subject to the license terms contained
# in the license file that is distributed with this file.

from setuptools import setup, find_packages
import codecs
import os

here = os.path.abspath(os.path.dirname(__file__))

with open("LIB_DESCRIPTION.md", "r") as fh:
    long_description = fh.read()

VERSION = '0.3.3'
DESCRIPTION = 'Toolkit for data science and machine learning built by Spotfire'

REQUIREMENTS = [
    'pandas==2.0.3',
    'scikit-learn==1.3.0',
    'tensorflow==2.13.0',
    'numpy==1.24.3',
    'fastapi==0.85.0',
    'uvicorn==0.18.3',
    'category_encoders==2.5.0',
    'matplotlib==3.7.2',
    'seaborn==0.12.2',
    'scikeras==0.11.0',
    'xgboost==1.7.6',
    'cloudpickle==2.2.1',
    'mlxtend==0.21.0',
    'supersmoother==0.4',
    'statsmodels==0.14.0',
    'nltk==3.8.1',
]

DEV_REQUIREMENTS = [
    'pytest==7.1.3',
    'pytest-cov==4.0.0',
    'pylint==2.15.3',
    'black==22.8.0',
    'mypy==0.982',
    'fire==0.4.0',
]

setup(
    name='spotfire-dsml',
    version=VERSION,
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://community.tibco.com/s/article/Python-toolkit-for-data-science-and-machine-learning-in-Spotfire',
    maintainer='Spotfire Python Package Support',
    maintainer_email='spotfirepython@tibco.com',
    license='BSD License (BSD 3-Clause License)',
    packages=find_packages(
        exclude=[
            'spotfire_dsml.modelling',
            'spotfire_dsml.exploration',
            'spotfire_dsml.module_template',
            'spotfire_dsml.utility'
        ]
    ),
    package_data={
        'PROJECT_NAME_URL': [
            'py.typed',
        ]
    },
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Operating System :: Unix",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ],
    install_requires=REQUIREMENTS,
    extras_require={
        'dev': DEV_REQUIREMENTS,
    },
    entry_points={
        'console_scripts': [
            'PROJECT_NAME_URL=project_name.my_module:main',
        ]
    },
    python_requires='>=3.8, <4',
)


from fastapi import APIRouter
from pydantic import BaseModel
from spotfire_dsml.time_series import preprocessing
import pandas as pd
import numpy as np

class MissingValueImputationRequest(BaseModel):
    ts: dict
    index: int
    numeric_method: str
    non_numeric_method: str
    numeric_constant: float
    non_numeric_constant: str
    warning_threshold: float
    n_previous: int

class ResamplingRequest(BaseModel):
    ts: dict
    index: int
    rule: str
    n_samples: int
    fill_method: str
    sort: bool

class MinMaxNormalizationRequest(BaseModel):
    ts: dict
    new_min: float
    new_max: float

class IndexNormalizationRequest(BaseModel):
    ts: dict
    index: int
    new_min: float
    new_max: float
    dt_format: str

router = APIRouter(
    prefix="/time_series",
    tags=["time_series"],
    responses={404: {"description": "Not found"}},
)


@router.post("/")
async def missing_value_imputation(missing_value_imputation_request: MissingValueImputationRequest):
    """
    Imputing time series for missing values.

    Parameters
    ----------
    missing_value_imputation_request: json structure with the following fields
    ts : data frame
        A data frame where a column is the timestamp or index and all of the
        other columns are readings of variables at corresponding timestamps.
        The number of columns can be an arbitrary number. The values can be
        numeric or non-numeric such as categorical (dictionary) or string. Required.
    index : str, int
        The index/timestamp column: if string it is interpreted as the name
        of the column. Value of -1 indicates that there is no index column and
        the function will create the index column. Required.
    numeric_method : str, int
        Type of imputation for numeric variables. Default = "previous".
        Options:
        0 or "nearest"
        1 or “constant” (constant specified with “numeric_constant”)
        2 or “previous”
        3 or “next”
        4 or “linear”
        5 or “mean”
        6 or “mean_over_n_previous”
        7 or “median_over_n_previous”
    non_numeric_method : str, int
        Type of imputation for non-numeric variables. Default = "previous".
        Options:
        0 or “nearest”
        1 or “constant”
        2 or “previous”
        3 or “next”
    numeric_constant : int, float, numeric
        The numeric constant to be used when the type of numeric imputation
        is set to “constant”. Ignored when “numeric_method” is not
        constant. Default = 0.
    non_numeric_constant : str
        The non-numeric constant to be used when the type of non-numeric
        imputation is set to “constant”. Ignored when “non_numeric_method”
        is not constant. Default = N/A.
    n_previous : int
        The number of past values to include in calculation when
        “mean_over_n_previous” or “median_over_n_previous” is set for the numeric
        imputation, ignored otherwise. Default = 3.
    warning_threshold : float
        A factor that describes the minimum fraction of missing values per column
        needed to invoke a warning to the user. Default = 0.5.

    Raises:
        ValueError: ValueError will be raised if the correlation method
                    specified is wrong or not enough features are selected

    Returns:
        pd.DataFrame: A dataframe with missing values imputed will be returned
    """
    df = pd.json_normalize(missing_value_imputation_request.data)

    result = preprocessing.missing_value_imputation(
        df,
        missing_value_imputation_request.index,
        missing_value_imputation_request.numeric_method or "previous",
        missing_value_imputation_request.non_numeric_method or "previous",
        missing_value_imputation_request.numeric_constant or 0,
        missing_value_imputation_request.non_numeric_constant or "N/A",
        missing_value_imputation_request.n_previous or 3,
        missing_value_imputation_request.warning_threshold or 0.5
    )

    return result
